/* hdparm.c - Command line interface to get/set hard disk parameters */
/*          - by Mark S. Lord (c) 1994 -- freely distributable */

#include <linux/hdreg.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <linux/fs.h>
#include <fcntl.h>
#include <errno.h>

char *progname;
struct hd_driveid id;
static long unmask = 0, readonly = 0, mult = 0, readahead = 0;
static int verbose = 0, get_identity = 0, get_geom = 0, get_size = 0;
static int set_unmask = 0, set_readonly = 0, set_mult = 0, set_readahead = 0;
static int get_unmask = 0, get_readonly = 0, get_mult = 0, get_readahead = 0;
static int flagcount = 0;

char *cfg_str[] =
{	"", " HardSect", " SoftSect", " NotMFM", " HdSw>15uSec", " SpinMotCtl",
	" Fixed", " Removeable", " DTR<=5Mbs", " DTR>5Mbs", " DTR>10Mbs",
	" RotSpdTol>.5%", " dStbOff", " TrkOff", " FmtGapReq", " nonMagnetic"
};

char *SlowMedFast[]	= {"slow", "medium", "fast"};
char *BuffType[]	= {"?", "1Sect", "DualPort", "DualPortCache"};

#define YN(b)	(((b)==0)?"no":"yes")

static void dmpstr (char *prefix, unsigned int i, char *s[], unsigned int maxi)
{
	printf(prefix);
	printf( (i > maxi) ? "?" : s[i] );
}

static void dump_identity (struct hd_driveid *id)
{
	int i;
	printf("\n Model=%.40s, FwRev=%.20s, SerialNo=%.8s", 
		id->model, id->fw_rev, id->serial_no);
	printf ("\n Config={");
	for (i=0; i<=15; i++) if (id->config & (1<<i)) printf (cfg_str[i]);
	printf (" }\n");
	printf (" Physical CHS=%d/%d/%d, TrkSize=%d, SectSize=%d, ECCbytes=%d\n",
		id->cyls, id->heads, id->sectors,
		id->track_bytes, id->sector_bytes, id->ecc_bytes);
	dmpstr (" BuffType=",id->buf_type,BuffType,3);
	printf (", BuffSize=%dKB, MaxMultSect=%d\n", id->buf_size/2, id->max_multsect);
	printf (" Features: DblWordIO=%s, LBA=%s, DMA=%s",
		YN(id->dword_io&1),YN(id->capability&2),YN(id->capability&1));
	dmpstr (", tPIO=",id->tPIO,SlowMedFast,2);
	if (id->capability&1)
		dmpstr (", tDMA=",id->tDMA,SlowMedFast,2);
	printf ("\n (%s): Logical CHS=%d/%d/%d, TotSect=%d",
		(id->cur_valid&1)?"valid":"maybe",
		id->cur_cyls, id->cur_heads, id->cur_sectors, *(int *)&id->cur_capacity0);
	if (id->capability&2)
		printf (", MaxLBAsect=%d", id->lba_capacity);
	printf("\n MultSect=");
	if (id->max_multsect && (id->multsect_valid&1))
		printf("%d", id->multsect);
	else
		printf("off");
	if (id->capability&1)
		printf (", DMA-1w=%04X, DMA-mw=%04X", id->dma_1word, id->dma_mword);
	printf ("\n\n");
}

void process_dev (char *devname)
{
	int fd, rc;
	static long parm;

	fd = open (devname, O_RDONLY);
	if (fd < 0) {
		printf ("could not open %s, errno=%d\n", devname, errno);
		exit(1);
	}

	printf("\n%s:\n", devname);
	if (set_readahead) {
		printf (" setting fs read-ahead to %ld", readahead);
		rc=ioctl (fd, BLKRASET, readahead);
		if (rc)	printf (" - BLKRASET failed, errno=%d\n", errno);
		else	printf ("\n");
	}
	if (set_mult) {
		printf (" setting mult-count to %ld", mult);
		rc=ioctl (fd, HDIO_SETMULTCOUNT, &mult);
		if (rc)	printf (" - HDIO_SETMULTCOUNT failed, errno=%d", errno);
		else {
			char buf[512];
			rc=ioctl (fd, BLKFLSBUF, NULL);
			if (rc) printf (" - BLKFLSBUF failed, errno=%d", errno);
			errno=read(fd, buf, sizeof(buf));
			if (rc == -1) printf (" - read failed, errno=%d", errno);
		}
		printf ("\n");
	}
	if (set_readonly) {
		printf (" setting read-only to %ld", readonly);
		rc=ioctl (fd, BLKROSET, &readonly);
		if (rc)	printf (" - BLKROSET failed, errno=%d\n", errno);
		else	printf ("\n");
	}
	if (set_unmask) {
		printf (" setting unmask to %ld", unmask);
		rc=ioctl (fd, HDIO_SETUNMASKINTR, &unmask);
		if (rc)	printf (" - HDIO_SETUNMASKINTR failed, errno=%d\n", errno);
		else	printf ("\n");
	}

	if (verbose || get_readahead || !flagcount) {
		rc=ioctl (fd, BLKRAGET, &parm);
		if (rc) printf (" BLKRAGET failed, errno=%d\n", errno);
		else    printf (" readahead = %ld\n", parm);
	}
	if (verbose || get_mult || !flagcount) {
		rc=ioctl (fd, HDIO_GETMULTCOUNT, &parm);
		if (rc) printf (" HDIO_GETMULTCOUNT failed, errno=%d\n", errno);
		else    printf (" multcount = %ld\n", parm);
	}
	if (verbose || get_readonly || !flagcount) {
		rc=ioctl (fd, BLKROGET, &parm);
		if (rc) printf (" BLKROGET failed, errno=%d\n", errno);
		else    printf (" readonly  = %ld\n", parm);
	}
	if (verbose || get_unmask || !flagcount) {
		rc=ioctl (fd, HDIO_GETUNMASKINTR, &parm);
		if (rc) printf (" HDIO_GETUNMASKINTR failed, errno=%d\n", errno);
		else    printf (" unmask    = %ld\n", parm);
	}
	if (verbose || get_geom || !flagcount) {
		static struct hd_geometry g;
		rc=ioctl (fd, HDIO_GETGEO, &g);
		if (rc) printf (" HDIO_GETGEO failed, errno=%d\n", errno);
		else    printf (" geometry  = %d/%d/%d, offset = %d\n",
				g.cylinders, g.heads, g.sectors, g.start);
	}
	if (verbose || get_size || !flagcount) {
		rc=ioctl (fd, BLKGETSIZE, &parm);
		if (rc) printf (" BLKGETSIZE failed, errno=%d\n", errno);
		else    printf (" sectors   = %ld\n", parm);
	}

	if (get_identity) {
		rc=ioctl (fd, HDIO_GETIDENTITY, &id);
		if (!rc)
			dump_identity(&id);
		else if (errno == -ENOMSG)
			printf (" no identification info available\n");
		else
			printf (" get-identity failed, errno=%d\n", errno);
	}
	close (fd);
}

void usage_error (void)
{
	fprintf(stderr,"%s - get/set hard drive parameters - version 1.1\n\n", progname);
	fprintf(stderr,"Usage:  %s  [options] [device] ..\n\n", progname);
	fprintf(stderr,"Options:\n");
	fprintf(stderr," -a 	get/set fs read-ahead\n");
	fprintf(stderr," -g	display drive geometry\n");
	fprintf(stderr," -h	display terse usage information\n");
	fprintf(stderr," -i	display boot-time drive identification\n");
	fprintf(stderr," -m   	get/set multiple sector count\n");
	fprintf(stderr," -r 	get/set read-only flag\n");
	fprintf(stderr," -s	display device size in sectors\n");
	fprintf(stderr," -u 	get/set unmask-irq flag\n");
	fprintf(stderr," -v	same as -agmrsu (default)\n\n");
	exit(1);
}

main(argc, argv)
int argc;
char **argv;
{
	char c, *p;

	if  ((progname = (char *) strrchr(*argv, '/')) == NULL)
                progname = *argv;
        else
                progname++;
	++argv;

	if (!--argc)
		usage_error();
	while (argc--) {
		p = *argv++;
		if (*p == '-') {
			if (!*++p)
				usage_error();
			while (c = *p++) {
				++flagcount;
				switch (c) {
					case 'v':
						verbose = 1;
						break;
					case 'i':
						get_identity = 1;
						break;
					case 'g':
						get_geom = 1;
						break;
					case 's':
						get_size = 1;
						break;
					case 'u':
						get_unmask = 1;
						if (!*p && argc && isdigit(**argv))
							p = *argv++, --argc;
						if (*p == '0' || *p == '1') {
							set_unmask = 1;
							unmask = *p++ - '0';
						}
						break;
					case 'r':
						get_readonly = 1;
						if (!*p && argc && isdigit(**argv))
							p = *argv++, --argc;
						if (*p == '0' || *p == '1') {
							set_readonly = 1;
							readonly = *p++ - '0';
						}
						break;
					case 'm':
						get_mult = 1;
						mult = 0;
						if (!*p && argc && isdigit(**argv))
							p = *argv++, --argc;
						while (isdigit(*p)) {
							set_mult = 1;
							mult *= 10;
							mult += *p++ - '0';
						}
						break;
					case 'a':
						get_readahead = 1;
						readahead = 0;
						if (!*p && argc && isdigit(**argv))
							p = *argv++, --argc;
						while (isdigit(*p)) {
							set_readahead = 1;
							readahead *= 10;
							readahead += *p++ - '0';
						}
						break;
					case 'h':
					default:
						usage_error();
				}
			}
			if (!argc)
				usage_error();
		} else {
			process_dev (p);
		} 
	}
	exit (0);
}
